;; -*- hen -*-
(declare (unit LWS-response)
		 (uses LWS-headers
			   uri)
		 (export <lws:response>
				 initialize

				 lws:info?
				 lws:success?
				 lws:redirect?
				 lws:error?
				 lws:client-error?
				 lws:server-error?

				 lws:code-of
				 lws:message-of
				 lws:headers-of
				 lws:content-of

				 lws:status-of ;; "200 OK" のような文字列を返す

				 lws:set-code!
				 lws:set-message!
				 lws:set-headers!
				 lws:set-content!))

(use tinyclos
	 format)

(include "cleited-macros.scm")

(define-class <lws:response> ()
  (code
   message
   headers
   content))

(define-accessors <lws:response>
  (code
   message
   headers
   content)
  prefix: lws:)


(define-method (initialize (this <lws:response>) args)
  (call-next-method)

  (define (init #!key (code 0) (message "") (headers (make <lws:headers>)) (content ""))

	(assert (integer?     code))
	(assert (string?      message))
	(assert (instance-of? headers <lws:headers>))
	(assert (string?      content))

	(initialize-slots this (list 'code    code
								 'message message
								 'headers headers
								 'content content)))

  (apply init args))


(define-method (lws:status-of (this <lws:response>))
  (with-slots (code message) this
	  (format #f "~3,'0,,D ~A" code message)))


(define-method (lws:info? (this <lws:response>))
  (with-slots (code) this
      (and (>= code 100)
		   (<  code 200))))

(define-method (lws:success? (this <lws:response>))
  (with-slots (code) this
      (and (>= code 200)
		   (<  code 300))))

(define-method (lws:redirect? (this <lws:response>))
  (with-slots (code) this
      (and (>= code 300)
		   (<  code 400))))

(define-method (lws:error? (this <lws:response>))
  (with-slots (code) this
      (and (>= code 400)
		   (<  code 600))))

(define-method (lws:client-error? (this <lws:response>))
  (with-slots (code) this
      (and (>= code 400)
		   (<  code 500))))

(define-method (lws:server-error? (this <lws:response>))
  (with-slots (code) this
      (and (>= code 500)
		   (<  code 600))))
