;; -*- hen -*-

(define-macro (with-mutex-locking mutex . body)
  (let ((mtx (gensym)))
	`(begin
	   (define ,mtx ,mutex)
	   (dynamic-wind
		   (lambda ()
			 (mutex-lock! ,mtx))

		   (lambda ()
			 ,@body)

		   (lambda ()
			 (mutex-unlock! ,mtx))))))

(define-macro (with-slots slots obj . body)
  ;; (with-slots (a b) this
  ;;    (+ a b))
  ;; ↓
  ;; (let ((a (slot-ref this 'a))
  ;;       (b (slot-ref this 'b)))
  ;;   (+ a b))
  (let ((this (gensym)))
	`(begin
	   (define ,this ,obj)

	   (let ,(map (lambda (slot)
					`(,slot (slot-ref ,this ',slot)))
				  slots)
		 ,@body))))

(define-macro (bind-exit var . body)
  ;; (bind-exit exit
  ;;    ...
  ;;    (exit 100)
  ;;    ...)
  ;; ↓
  ;; (call/cc
  ;;   (lambda (exit)
  ;;      ...
  ;;      (exit 100)
  ;;      ...))
  `(call/cc (lambda (,var)
			  ,@body)))

(define-macro (define-accessors class slots #!key prefix)
  ;; (define-accessors <uri> (scheme host ...) prefix: uri)
  ;; ↓
  ;; (begin
  ;;   (define-method (uri:scheme-of (this <uri>))
  ;;       (slot-ref this 'scheme))
  ;;   (define-method (uri:set-scheme! (this <uri>) scheme)
  ;;       (slot-set! this 'scheme scheme))
  ;;   ...)
  (define (make-getter slot)
	(let ((getter-sym
		   (string->symbol (sprintf "~a~a-of" (if prefix
												  prefix "") slot))))
	  `(define-method (,getter-sym (this ,class))
		 (slot-ref this ',slot))))

  (define (make-setter slot)
	(let ((setter-sym
		   (string->symbol (sprintf "~aset-~a!" (if prefix
													prefix "") slot))))
	  `(define-method (,setter-sym (this ,class) ,slot)
		 (slot-set! this ',slot ,slot))))

  `(begin
	 ,@(map make-getter slots)
	 ,@(map make-setter slots)))
