;; -*- hen -*-
(declare (unit cleited-utils)
		 (uses uri)
		 (export util:clean-uri
				 util:rfc3339->epoch
				 util:sqlite-quote
				 util:epoch->gmt-string
				 util:get-config
				 util:set-config!
                 util:xml->utf8-xml
                 util:utf8->unicode
                 util:unicode->utf8
                 util:escape-xml
                 util:sxml->xml
                 util:cat-normal
                 util:cat-clark))

(use posix
	 tinyclos
	 rfc3339
     (srfi 1)
	 (srfi 13)
	 regex
	 format
	 ssax
	 sxml-tools
	 iconv
	 sqlite3)

(include "cleited-macros.scm")


(define (util:clean-uri uri)
  (let ((new-uri (make <uri> uri)))

	(uri:canonicalize! new-uri)
	(uri:set-fragment! new-uri #f)

	new-uri))

(define (util:rfc3339->epoch dt)
  (assert (rfc3339-record? dt))

  ;; timezone を指定しても無視されてしまふので、後から補正する。
  (define utc-epoch
	(local-time->seconds (vector (rfc3339-record:second dt)
								 (rfc3339-record:minute dt)
								 (rfc3339-record:hour   dt)
								 (rfc3339-record:mday   dt)
								 (- (rfc3339-record:month  dt)    1)
								 (- (rfc3339-record:year   dt) 1900)
								 0  ;; wday
								 0  ;; yday
								 #f ;; dstflag
								 0  ;; timezone
								 )))

  (define offsetmin (rfc3339-record:offsetmin dt))

  (if offsetmin
	  (- utc-epoch
		 (* 60 offsetmin))
							 
	  utc-epoch))


(define (util:sqlite-quote str #!key (char #\'))
  
  (define (escape ch)
	(if (char=? ch char)
		(string char char)
		(string ch)))

  (string-concatenate (append (list (string char))
                              (map escape (string->list str))
                              (list (string char)))))


(define (util:epoch->gmt-string epoch)
  ;; UTC と GMT の間には閏秒分の誤差があるが、まあ氣にしなくても良いだらう。
  ;; それもこれも私の開發環境で srfi-19.egg が動かないからだ。
  (define utc (seconds->utc-time epoch))

  ;; 暗號だ…
  (format #f "~A, ~2,'0,,D ~A ~4,'0,,D ~2,'0,,D:~2,'0,,D:~2,'0,,D GMT"
		  (list-ref '(Sun Mon Tue Wed Thu Fri Sat) (vector-ref utc 6)) ;; wday
		  (vector-ref utc 3) ;; mday
		  (list-ref '(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec) (vector-ref utc 4)) ;; month
		  (+ 1900 (vector-ref utc 5)) ;; year
		  (vector-ref utc 2) ;; hour
		  (vector-ref utc 1) ;; minute
		  (vector-ref utc 0))) ;; second
  

(define (util:get-config db key)
  ;; 設定値が無い場合は #f を返す。
  (condition-case
   (sqlite3:first-result db (string-append "SELECT conf_val    "
										   "  FROM config      "
										   " WHERE conf_key = ?")
						 (->string key))
   [(exn sqlite3)
	#f]))


(define (util:set-config! db key value)
  (sqlite3:exec db (string-append "REPLACE INTO config         "
								  "        (conf_key, conf_val)"
								  " VALUES (?       , ?       )")
				(->string key)
                (->string value)))


(define (util:xml-numref->utf8 src)
  ;; &#DDDDD; => 10進の參照
  ;; &#xHHHH; => 16進の參照
  ;;
  ;; 但し變換結果が < > & になった場合はそれぞれ實體參照にする。
  ;;
  (let* [(re-numref (regexp "&#((\\d+)|x([0-9a-fA-F]+));"))
         (fix       (lambda (str)
                      (cond [(equal? str "<") "&lt;" ]
                            [(equal? str ">") "&gt;" ]
                            [(equal? str "&") "&amp;"]
                            [(equal? str "\"") "&quot;"]
                            [(equal? str "\'") "&apos;"]
                            [else             str    ])))
         (decode    (rec (decode src)
                         (let [(m (string-search-positions re-numref src))]
                           (if m
                               ;; マッチした
                               (let [(whole   (car m))
                                     (decimal (caddr m))
                                     (hex     (cadddr m))]
                                
                                 (string-append (substring src 0 (car whole))
                                                (fix (cond [decimal
                                                            (util:unicode->utf8
                                                             (string->number
                                                              (substring src (car decimal) (cadr decimal))
                                                              10))]

                                                           [hex
                                                            (util:unicode->utf8
                                                             (string->number
                                                              (substring src (car hex) (cadr hex))
                                                              16))]

                                                           [else
                                                            (error "internal error")]))
                                                (decode
                                                    (substring src (cadr whole)))))
                               ;; マッチしなかった
                               src))))]
    (decode src)))
                      

(define (util:xml->utf8-xml src)
  (define encoding
	(bind-exit exit
      
	    ((SSAX:make-parser
		     NEW-LEVEL-SEED
			 (lambda (name attr ns expect seed)
			   seed)

			 FINISH-ELEMENT
			 (lambda (name attr ns parent-seed seed)
			   seed)

			 CHAR-DATA-HANDLER
			 (lambda (s1 s2 seed)
			   seed)

			 PI
			 ((xml . (lambda (port tag seed)
					   (let* [(sxml (SSAX:XML->SXML
								 	    (open-input-string
									        (string-append "<pi "
														   (SSAX:read-pi-body-as-string port)
														   " />"))
										'()))
							  (enc  ((sxpath "/pi/@encoding/text()") sxml))]

						 (unless (null? enc)
						   (exit (car enc))))))))
		 
		 (open-input-string src) #f)

		;; エンコーディング指定が無ければ UTF-8 と見做す。
		"UTF-8"))

  (util:xml-numref->utf8
      (iconv (iconv-open "UTF-8" encoding)
             src)))


(define (util:utf8->unicode src)
  
  (define (decode l)
    (if (null? l)
        l

        (cond [(= 0 (bitwise-and (car l) #x80))
               ;; １バイト
               (cons (car l)
                     (decode (drop l 1)))]

              [(= #xC0 (bitwise-and (car l) #xE0))
               ;; ２バイト
               (cons (bitwise-ior (arithmetic-shift (bitwise-and (list-ref l 0) #x1F) 6)
                                  (bitwise-and (list-ref l 1) #x3F))
                     (decode (drop l 2)))]

               [(= #xE0 (bitwise-and (car l) #xF0))
                ;; ３バイト
                (cons (bitwise-ior (arithmetic-shift (bitwise-and (list-ref l 0) #x0F) 12)
                                   (arithmetic-shift (bitwise-and (list-ref l 1) #x3F)  6)
                                   (bitwise-and (list-ref l 2) #x3F))
                      (decode (drop l 3)))]

               [(= #xF0 (bitwise-and (car l) #xF8))
                 ;; ４バイト
                (cons (bitwise-ior (arithmetic-shift (bitwise-and (list-ref l 0) #x07) 18)
                                   (arithmetic-shift (bitwise-and (list-ref l 1) #x3F) 12)
                                   (arithmetic-shift (bitwise-and (list-ref l 2) #x3F)  6)
                                   (bitwise-and (list-ref l 3) #x3F))
                      (decode (drop l 4)))]

               [(= #xF8 (bitwise-and (car l) #xFC))
                ;; ５バイト
                (cons (bitwise-ior (arithmetic-shift (bitwise-and (list-ref l 0) #x03) 24)
                                   (arithmetic-shift (bitwise-and (list-ref l 1) #x3F) 18)
                                   (arithmetic-shift (bitwise-and (list-ref l 2) #x3F) 12)
                                   (arithmetic-shift (bitwise-and (list-ref l 3) #x3F)  6)
                                   (bitwise-and (list-ref l 4) #x3F))
                      (decode (drop l 5)))]

               [(= #xFC (bitwise-and (car l) #xFE))
                ;; ６バイト
                (cons (bitwise-ior (arithmetic-shift (bitwise-and (list-ref l 0) #x01) 30)
                                   (arithmetic-shift (bitwise-and (list-ref l 1) #x3F) 24)
                                   (arithmetic-shift (bitwise-and (list-ref l 2) #x3F) 18)
                                   (arithmetic-shift (bitwise-and (list-ref l 3) #x3F) 12)
                                   (arithmetic-shift (bitwise-and (list-ref l 4) #x3F)  6)
                                   (bitwise-and (list-ref l 5) #x3F))
                      (decode (drop l 6)))])))

  (list->string (map integer->char
                     (decode (map char->integer
                                  (string->list src))))))


(define (util:unicode->utf8 src)
  (cond [(char? src)
         
         (let [(c (char->integer src))]
           (cond ((<= c #x7F)
                  ;; １バイト
                  (string (integer->char (bitwise-and c #x7F))))

                 ((<= c #x07FF)
                  ;; ２バイト
                  (string (integer->char (bitwise-ior #xC0 (bitwise-and #x1F (arithmetic-shift c -6))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F c)))))

                 ((<= c #xffff)
                  ;; ３バイト
                  (string (integer->char (bitwise-ior #xE0 (bitwise-and #x0F (arithmetic-shift c -12))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c  -6))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F c)))))

                 ((<= c #x1fffff)
                  ;; ４バイト
                  (string (integer->char (bitwise-ior #xF0 (bitwise-and #x07 (arithmetic-shift c -18))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c -12))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c  -6))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F c)))))

                 ((<= c #x3ffffff)
                  ;; ５バイト
                  (string (integer->char (bitwise-ior #xF8 (bitwise-and #x03 (arithmetic-shift c -24))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c -18))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c -12))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c  -6))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F c)))))

                 ((<= x #x7fffffff)
                  ;; ６バイト
                  (string (integer->char (bitwise-ior #xFC (bitwise-and #x01 (arithmetic-shift c -30))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c -24))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c -18))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c -12))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F (arithmetic-shift c  -6))))
                          (integer->char (bitwise-ior #x80 (bitwise-and #x3F c)))))

                 (else
                  (error (sprintf "can't encode to UTF-8: %d" c)))))]

        [(integer? src)
         (util:unicode->utf8 (integer->char src))]

        [else
         (string-concatenate (map util:unicode->utf8 (string->list src)))]))


(define (util:escape-xml src)
  
  (define (encode c)
    (cond [(eq? c #\<) "&lt;"]
          [(eq? c #\>) "&gt;"]
          [(eq? c #\&) "&amp;"]
          [(eq? c #\") "&quot;"]
          [(eq? c #\') "&apos;"]
          [else        (list->string (list c))]))
  
  (string-concatenate (map encode (string->list src))))
           


;; sxml->xml (borrowed from http://homepage1.nifty.com/blankspace/scheme/nsx.html)
(define (util:sxml->xml sxml)
  (define (make-attr l ac)
    (if (null? l) ac
        (make-attr
         (cdr l) (string-append ac " " (symbol->string (caar l))
                                "=\"" (util:escape-xml (cadar l)) "\""))))
  (define (make-xml x bef aft)
    (cond ((string? x) (append bef (list (util:escape-xml x)) aft))
          ((and (pair? x) (symbol? (car x)))
           (let ((tag (symbol->string (car x))))
             (if (and (not   (null? (cdr x)))
                      (pair? (cadr x))
                      (eq? (caadr  x) '@))
                 (make-xml (cddr x)
                           (append bef (list 'BO tag (make-attr (cdadr x) "") 'BC))
                           (append     (list 'EO tag 'EC) aft))
                 (make-xml (cdr x)
                           (append bef (list 'BO tag 'BC))
                           (append     (list 'EO tag 'EC) aft)))))
          ((pair? x) (make-xml '()
                               (make-xml (car x) bef '())
                               (make-xml (cdr x) '() aft)))
          (else (append bef aft))))
  ;; main
  (make-xml sxml '() '()))

;; utility
(define (util:cat-normal l s)
  (cond ((null? l) s)
        ((string? (car l))
         (util:cat-normal (cdr l) (string-append s (car l))))
        (else (util:cat-normal (cdr l)
                               (string-append s (case (car l)
                                                  ((BO)    "<")
                                                  ((EO)    "</")
                                                  ((BC EC) ">\n")))))))
;; utility
(define (util:cat-clark l s i)
  (cond ((null? l) s)
        ((string? (car l))
         (util:cat-clark (cdr l) (string-append s (car l)) i))
        (else
         (case (car l)
           ((BO)    (util:cat-clark (cdr l) (string-append s "<" ) (+ i 1)))
           ((EO)    (util:cat-clark (cdr l) (string-append s "</") (- i 1)))
           ((BC EC) (util:cat-clark (cdr l)
                                    (string-append s "\n" (make-string i #\ ) ">") i))))))