;; -*- hen -*-
(declare (unit config)
         (uses cleited-passive-node
               cleited-utils
               account)
         (export config:install-methods!))

(use tinyclos
     sqlite3)

(define (config:install-methods! node)
  (assert (instance-of? node <cleited-passive-node>))

  (wrap-method! node 'config:list config:list)
  (wrap-method! node 'config:set  config:set ))


(define (config:list arg db-pod)
  (let [(auth (get-arg arg 'auth check: list?))]

    (account:check auth db-pod need-root: #t)

    (db-pod:run db-pod
        (lambda (db)
          (sqlite3:map-row
              (lambda (key value type description)
                `((key         . ,(string->symbol key))
                  (value       . ,(cond [(or (equal? type "natural")
                                             (equal? type "integer")
                                             (equal? type "real"   ))
                                         (string->number value)]

                                        [(equal? type "boolean")
                                         (cond [(equal? value "true")
                                                'true]

                                               [(equal? value "false")
                                                'false]

                                               [else
                                                (error "not a boolean: ~a" value)])]

                                        [(equal? type "string")
                                         value]

                                        [else
                                         (error "unknown type: ~a" type)]))
                  (type        . ,(string->symbol type))
                  (description . ,description)))
              db
              (string-append "SELECT conf_key, conf_val, conf_type, description"
                             "  FROM config                                    "))))))


(define (config:set arg db-pod)
  (let [(auth  (get-arg arg 'auth  check: list?))
        (key   (get-arg arg 'key   check: symbol?))
        (value (get-arg arg 'value))]

    (account:check auth db-pod need-root: #t)

    (db-pod:run db-pod
        (lambda (db)
          (sqlite3:exec db "BEGIN IMMEDIATE")

          (let* [(type
                    (condition-case
                     (sqlite3:first-result
                        db
                        "SELECT conf_type FROM config WHERE conf_key = ?"
                        (->string key))
                     [(exn sqlite3)
                      (error (sprintf "Unknown config key: ~a" key))]))
                
                 (valid?
                    (cond [(equal? "natural" type)
                           (lambda (n)
                             (and (integer? n)
                                  (> n 0)))]

                          [(equal? "integer" type)
                           integer?]

                          [(equal? "real" type)
                           number?]

                          [(equal? "boolean" type)
                           (lambda (b)
                             (or (eq? b 'true)
                                 (eq? b 'false)))]

                          [(equal? "string" type)
                           string?]

                          [else
                           (error (sprintf "Unknown type: ~a" type))]))]
            
            (unless (valid? value)
              (error (sprintf "Invalid type of value `~a' for key `~a'."
                              value
                              key)))

            (sqlite3:exec db (string-append "UPDATE config      "
                                            "   SET conf_val = ?"
                                            " WHERE conf_key = ?")
                          (->string value)
                          (->string key)))

          (sqlite3:exec db "COMMIT")))
    'ok))