;; -*- hen -*-
(declare (unit db-pod)
		 (export <db-pod>
				 initialize
				 db-pod:run))

(use tinyclos
	 srfi-18
	 sqlite3)

(define-class <db-pod> ()
  (dbh      ;; <sqlite3:database>
   in-use   ;; <boolean>
   mutex    ;; mutex
   released ;; condition-variable
   ))


(define-method (initialize (this <db-pod>) args)
  (call-next-method)

  (define (init #!key database)
	(assert (instance-of? database <sqlite3:database>))

	(initialize-slots this
					  (list 'dbh      database
							'in-use   #f
							'mutex    (make-mutex)
							'released (make-condition-variable))))
  (apply init args))


(define-method (db-pod:run (this <db-pod>)
						   (proc <procedure>))

  (define dbh      (slot-ref this 'dbh     ))
  (define mutex    (slot-ref this 'mutex   ))
  (define released (slot-ref this 'released))

  (dynamic-wind
	  (lambda ()
		(mutex-lock! mutex))

	  (lambda ()
		;; dbh が使用中であれば、使用中でなくなるまで待つ。
		(let loop ()
		  (when (slot-ref this 'in-use)
			(mutex-unlock! mutex released)
			(mutex-lock! mutex)
			(loop)))

		;; 使用開始。使用中はロックを外して置く。
		(dynamic-wind
			(lambda ()
			  (slot-set! this 'in-use #t)
			  (mutex-unlock! mutex))

			(lambda ()
			  (proc dbh))

			(lambda ()
			  (mutex-lock! mutex)
			  (slot-set! this 'in-use #f)
			  ;; トランザクション中に死んだ可能性があるのでロールバック
			  ;; を試みる。
			  (condition-case
			   (sqlite3:exec dbh "ROLLBACK")

			   [(exn sqlite3) #f]))))

	  (lambda ()
		(mutex-unlock! mutex)
		(condition-variable-signal! released))))
