;; -*- hen -*-
(declare (unit feed-item)
		 (uses cleited-passive-node
			   cleited-utils
			   uri
			   account)
		 (export feed-item:install-methods!))

(use tinyclos
	 sqlite3)


(define (feed-item:install-methods! node)
  (assert (instance-of? node <cleited-passive-node>))

  (wrap-method! node 'feed-item:get      feed-item:get )
  (wrap-method! node 'feed-item:list     feed-item:list)
  (wrap-method! node 'feed-item:set-read feed-item:set-read))


(define (feed-item:get arg db-pod)
  (let [(auth     (get-arg arg 'auth     check: list?  ))
		(feed-uri (get-arg arg 'feed-uri check: string?))
		(item-uri (get-arg arg 'item-uri check: string?))]

	(account:check auth db-pod)

	(set! feed-uri (util:clean-uri feed-uri))
	(set! item-uri (uri:canonicalize (make <uri> item-uri)))

	(define item-list (feed-item:list `((auth      . ,auth                  )
										(feed-uri  . ,(uri->string feed-uri))
										(need-desc . true                   )
                                        (item-uri  . ,(uri->string item-uri)))
									  db-pod))
	
	(if (null? item-list)
		(error (sprintf (string-append "You have not subscribed to `~a', "
									   "or the feed does not have an item `~a'.")
						(uri->string feed-uri)
						(uri->string item-uri))))

	(car item-list)))
				 

(define (feed-item:list arg db-pod)
  (let [(auth      (get-arg arg 'auth      check: list?                      ))
		(feed-uri  (get-arg arg 'feed-uri  check: string?                    ))
		(need-desc (get-arg arg 'need-desc check: true-or-false? optional: #t))
        (item-uri  (get-arg arg 'item-uri  check: string?        optional: #t))]

	(account:check auth db-pod)

	(set! feed-uri (util:clean-uri feed-uri))

    (define item-uri-sql (if item-uri
                             (sprintf "AND I.item_uri = ~a"
                                      (util:sqlite-quote item-uri char: #\"))
                             ""))
    
	(db-pod:run db-pod
        (lambda (db)
		  (sqlite3:map-row
		      (lambda (feed-uri item-uri title description date read-before)
				`((feed-uri    . ,feed-uri                             )
				  (item-uri    . ,item-uri                             )
				  (title       . ,title                                )
				  ,@(if need-desc
						`((description . ,description))
						'())
				  (date        . ,date                                 )
				  (read-before . ,(number->true-or-false read-before))))
			  db
			  (string-append "SELECT F.feed_uri    ,                       "
							 "       I.item_uri    ,                       "
							 "       I.title       ,                       "
							 "       I.description ,                       "
							 "       I.date        ,                       "
							 "       EXISTS(SELECT *                       "
							 "                FROM read_items              "
							 "               WHERE item_id = I.item_id     "
							 "                 AND user_id = S.user_id)    "
							 "  FROM items AS I                            "
							 "  LEFT JOIN subscription AS S USING (feed_id)"
							 "  LEFT JOIN feeds        AS F USING (feed_id)"
							 " WHERE S.user_id = ? AND F.feed_uri = ?      "
                             "     " item-uri-sql                           )
			  (get-arg auth 'user-id)
			  (uri->string feed-uri))))))


(define (feed-item:set-read arg db-pod)
  (let [(auth        (get-arg arg 'auth        check: list?         ))
		(feed-uri    (get-arg arg 'feed-uri    check: string?       ))
		(item-uri    (get-arg arg 'item-uri    check: string?       ))
		(read-before (get-arg arg 'read-before check: true-or-false?))]

	(account:check auth db-pod)

	(set! feed-uri (util:clean-uri feed-uri))
	(set! item-uri (uri:canonicalize (make <uri> item-uri)))

	(db-pod:run db-pod
	    (lambda (db)
		  (sqlite3:exec db "BEGIN IMMEDIATE")

		  (condition-case
		   (let [(item-id (sqlite3:first-result
						   db (string-append "SELECT I.item_id                                          "
											 "  FROM items             AS I                             "
											 "  LEFT JOIN feeds        AS F USING (feed_id)             "
											 "  LEFT JOIN subscription AS S USING (feed_id)             "
											 " WHERE F.feed_uri = ? AND I.item_uri = ? AND S.user_id = ?")
						   (uri->string feed-uri)
						   (uri->string item-uri)
						   (get-arg auth 'user-id)))]
			 (if (true-or-false->boolean read-before)
				 ;; フラグを立てる
				 (sqlite3:exec db (string-append "INSERT OR IGNORE INTO read_items"
												 "       (item_id, user_id)       "
												 "VALUES (?      , ?      )       ")
							   item-id
							   (get-arg auth 'user-id))
				 ;; フラグを消す
				 (sqlite3:exec db (string-append "DELETE FROM read_items            "
												 " WHERE item_id = ? AND user_id = ?")
							   item-id
							   (get-arg auth 'user-id))))
		   [(exn sqlite3)
			(error (sprintf (string-append "You have not subscribed to `~a', "
										   "or the feed does not have an item `~a'.")
							(uri->string feed-uri)
							(uri->string item-uri)))])

		  (sqlite3:exec db "COMMIT")))
	
	'ok))
