;; -*- hen -*-
(declare (unit feed-poller)
		 (uses throttle
			   db-pod
			   feed
			   logger)
		 (export <feed-poller>
				 initialize
				 start!))

(include "cleited-macros.scm")

(use tinyclos
	 sqlite3)


(define-class <feed-poller> ()
  (db-pod          ;; 使用する db-pod
   throttle        ;; 使用する throttle
   min-interval    ;; 何分置きに poll すべきフィードを探すか
   poller-thread)) ;; poll すべきフィードを探すスレッド

(define-method (initialize (this <feed-poller>) args)
  (call-next-method)

  (define (init #!key db-pod throttle (min-interval-minute 5))
	
	(assert (instance-of? db-pod <db-pod>))
	(assert (instance-of? throttle <throttle>))
	(assert (integer?     min-interval-minute  ))

	(initialize-slots this
					  (list 'db-pod        db-pod
							'throttle      throttle
							'min-interval  min-interval-minute
							'poller-thread (make-thread
											(lambda ()
											  (poller-main this))))))

  (apply init args))


(define-method (start! (this <feed-poller>))
  ;; 別スレッドで feed-poller を開始する。
  (with-slots (poller-thread) this

      (thread-start! poller-thread)))


(define-method (poller-main (this <feed-poller>))
  (with-slots (db-pod min-interval) this
      (let loop ()
		;; poll 時刻を迎へてゐる feed があれば、それを別スレッドで
		;; poll する。
		(db-pod:run db-pod
		    (lambda (db)
			  (sqlite3:map-row (lambda (uri)
								 (thread-start! (make-thread (lambda ()
															   (child-main this uri)))))
							   db
							   (string-append "SELECT feed_uri         "
											  "  FROM feeds            "
											  " WHERE lastcheck + ? < ?")
							   (* 60 (string->number
									  (util:get-config db 'polling-interval-minute)))
							   (current-seconds))))

		;; min-interval 分だけ待つ
		(thread-sleep! (* 60 min-interval))
		
		(loop))))


(define-method (child-main (this <feed-poller>)
						   (uri  <string>))
  
  (with-slots (db-pod throttle) this
      (throttle:enqueue throttle
	       (lambda ()

			 ;(logger 'feed-poller 'debug
			;		 "started checking for `~a'..." uri)

			 (condition-case
			  (begin
				(feed:check! uri db-pod)
				;(logger 'feed-poller 'info
					;	"finished checking for `~a'" uri))
                )

			  [exn ()
				   (logger 'feed-poller 'error
						   "aborted checking for `~a' because of an error:~%~a"
						   uri
						   (with-output-to-string
							 (lambda ()
							   (print-error-message exn)
							   (print-call-chain))))])))))

