;; -*- hen -*-
(declare (unit logger)
		 (export logger
                 logger:set-threshold!
				 logger:set-file!))

(use mailbox
	 format
     srfi-1
	 srfi-18
	 posix)

;; none は threshold 指定でのみ使用される。(logger) へ渡す事は出來ない。
(define-constant logger:levels
  '(none error warn info debug))

(define logger:thread    #f)
(define logger:queue     #f)
(define logger:port      #f)
(define logger:threshold 'none)


(define (logger:level->int level)
  (let [(idx (list-index (lambda (sym)
                           (eq? sym level))
                         logger:levels))]
    (assert idx)
    idx))

(define (logger:setup)
  ;; logger スレッドが存在しなければ、まず queue を作ってからスレッドを
  ;; 起動。
  (unless logger:thread
	(set! logger:queue (make-mailbox))
	(set! logger:thread (make-thread logger:thread-main))

	(thread-start! logger:thread)))


(define (logger:thread-main)
  (let loop ()
	(let* [(command (mailbox-receive! logger:queue))
		   (type    (cdr (assq 'type command)))]

	  (cond [(eq? type 'set-port!)
			 (set! logger:port (cdr (assq 'port command)))]

            [(eq? type 'set-threshold!)
             (set! logger:threshold (cdr (assq 'level command)))]

			[(eq? type 'log)
             (call/cc (lambda (cancel)
                        (unless logger:port
                          ;; 出力ポート未設定
                          (fprintf (current-error-port)
                                   "Warning: call logger:set-file! before logging anything.~%~a~%"
                                   (cdr (assq 'message command)))
                          (cancel #f))

                        (when (< (logger:level->int logger:threshold)
                                 (logger:level->int (cdr (assq 'level command))))
                          ;; レベル閾値による抑壓
                          (cancel #f))

                        ;; 出力
                        (fprintf logger:port "~a~%" (cdr (assq 'message command)))
                        (flush-output logger:port)))]

			[else
			 (fprintf (current-error-port)
					  "Warning: unknown logger command type: ~a~%" type)]))
	(loop)))


(define (logger:set-threshold! threshold)
  (assert (memq threshold logger:levels))
  (logger:setup)
  (mailbox-send! logger:queue
                 `((type  . set-threshold!)
                   (level . ,threshold    ))))


(define (logger:set-file! fpath)
  (logger:setup)
  (mailbox-send! logger:queue
				 `((type . set-port!)
				   (port . ,(open-output-file fpath #:append)))))


(define (logger:format category format-str args)
  
  (define local-time
	(seconds->local-time (current-seconds)))
  
  (apply format
		 #f (string-append
			 "[~4,'0,,D-~2,'0,,D-~2,'0,,D ~2,'0,,D:~2,'0,,D:~2,'0,,D][~A] "
			 format-str)
		 
		 (+ 1900 (vector-ref local-time 5))
		 (+    1 (vector-ref local-time 4))
		 (vector-ref local-time 3)

		 (vector-ref local-time 2)
		 (vector-ref local-time 1)
		 (vector-ref local-time 0)

		 category
		 args))
  
		 
(define (logger category level format-str #!rest args)
  (assert (memq level logger:levels))
  (logger:setup)
  (mailbox-send! logger:queue
				 `((type    . log)
                   (level   . ,level)
				   (message . ,(logger:format category format-str args)))))
