;; -*- hen -*-
(declare (uses db-setup
			   logger
			   cleited-passive-node
			   cleited-constants
			   db-pod
			   throttle
			   feed-poller
			   account
			   feed
			   feed-item
			   opml
               config))

(include "cleited-macros.scm")
(use args
	 (srfi 18))

(define opt:home          cleited:default-home)
(define opt:port          cleited:default-port)
(define opt:user          cleited:default-user)
(define opt:group         cleited:default-group)
(define opt:client-limit  cleited:default-client-limit)
(define opt:poller-limit  cleited:default-poller-limit)
(define opt:logging-level cleited:default-logging-level)
(define opt:action        'start)

(define opts
  (list (args:make-option (h help) #:none
						  "display this text"
						  (usage))

		;; 制御項目
		(args:make-option (H home) (required: "DIR")
						  (sprintf "specify the home directory [~a]" opt:home)
						  (set! opt:home arg))
		
		(args:make-option (P port) (required: "PORT")
						  (sprintf "specify which port to use [~a]" opt:port)
						  (set! opt:port arg))

		(args:make-option (U user) (required: "USER")
						  (sprintf "specify which user to setuid [~a]" opt:user)
						  (set! opt:user arg))

		(args:make-option (G group) (required: "GROUP")
						  (sprintf "specify which group to setgid [~a]" opt:group)
						  (set! opt:group arg))

		(args:make-option (client-limit) (required: "LIMIT")
						  (sprintf "specify the limit of concurrency for serving clients [~a]"
                                   opt:client-limit)
						  (set! opt:client-limit (string->number arg)))

		(args:make-option (poller-limit) (required: "LIMIT")
						  (sprintf "specify the limit of concurrency for polling feeds [~a]"
                                   opt:poller-limit)
						  (set! opt:poller-limit (string->number arg)))

        (args:make-option (logging-level) (required: "LV")
                          (sprintf "specify the logging level (none/error/warn/info/debug) [~a]"
                                   opt:logging-level)
                          (set! opt:logging-level (string->symbol arg)))
		
		;; 實行項目
		(args:make-option (s stop) #:none
						  "Stop the currently running daemon gracefully"
						  (set! opt:action 'stop))

		(args:make-option (kill) #:none
						  "Kill the currently running daemon forcefully"
						  (set! opt:action 'kill))

		(args:make-option (r restart) #:none
						  "Restart the currently running daemon gracefully"
						  (set! opt:action 'restart))

		(args:make-option (f no-fork) #:none
						  "Debug the daemon by running without fork(2)ing"
						  (set! opt:action 'no-fork))))

(define (usage)
  (with-output-to-port (current-error-port)
	(lambda ()
	  (print "Usage: " (car (argv)) " [options]")
	  (newline)
	  (print (args:usage opts))))
  (exit 1))


(define (recall-running-pid)
  ;; 以下の二つの條件が滿されてゐる場合は、ノードが既に存在してゐるもの
  ;; と判斷する。
  ;; [1] cleited.pid が存在する
  ;; [2] cleited.pid がその PID によってロックされてゐる
  (bind-exit exit
      (define pid-file-name (sprintf "~a/cleited.pid" opt:home))

	  (when (file-exists? pid-file-name)
		;; pid ファイルが存在する
		(with-input-from-file pid-file-name
		  (lambda ()
			(define pid (read))

			(let [(locker (file-test-lock (current-input-port)))]
			  (when (and locker
						 (= locker pid))
				;; ファイルに書かれてゐるものと同じ pid がこのファイル
				;; をロックしてゐるので、pid の内容が正しいものと見做す。
				(exit pid))))))

	  ;; pid ファイルが無いか、内容が正しくない。(既に古くなってゐる？)
	  #f))

(define pidfile-port #f)
(define (leave-running-pid)
  
  (define pid-file-name (sprintf "~a/cleited.pid" opt:home))
  
  ;; pid ファイルを作り、ロックする
  (when (file-exists? pid-file-name)
	(delete-file pid-file-name))

  ;; この關數を抜けてもロックが外れないやうにする。
  (set! pidfile-port (open-output-file pid-file-name))
  
  (file-lock/blocking         pidfile-port)
  (write (current-process-id) pidfile-port)
  (flush-output               pidfile-port))
			

(define (start-node)
  (bind-exit exit
	  ;; pid 關係の處理
	  (let [(pid (recall-running-pid))]
		(when pid
		  (printf "Another instance of the CleiteD is already running with pid ~a~%"
				  pid)
		  (exit 1)))

	  (leave-running-pid)

	  (logger:set-file! (sprintf "~a/cleited.log" opt:home))
      (logger:set-threshold! opt:logging-level)
	  (logger 'main 'info "CleiteD started running with pid ~d." (current-process-id))

	  (let* [(dbh      (sqlite3:open (sprintf "~a/cleited.db" opt:home)))
			 (pod      (make <db-pod> database: dbh))]

		;; DB の準備
		(db-setup dbh)
		
		;; ノードを起動
		(let [(node (make <cleited-passive-node>
					  max-requests: opt:client-limit
					  db-pod:       pod
					  port:         opt:port))]

		  (account:install-methods!   node)
		  (feed:install-methods!      node)
		  (feed-item:install-methods! node)
		  (opml:install-methods!      node)
          (config:install-methods!    node)
		  
		  (start! node))
	  
		;; poller を起動
		(let* [(throttle (make <throttle>
						   concurrency: opt:poller-limit))
			   (poller (make <feed-poller>
						 throttle: throttle
						 db-pod:   pod))]

		  (start! poller))

	  
		;; 無限に停止
		(let loop ()
		  (thread-sleep! 1000)
		  (loop)))))


(define (daemonize-node)
  (process-fork (lambda () (start-node))))


(define (stop-node #!key force)
  (bind-exit exit
     (let [(pid (recall-running-pid))]

	   (unless pid
		 (print "No CleiteD is currently running.")
		 (exit 0))
	
	   (process-signal pid (if force
							   signal/kill
							   signal/quit))
	   (printf "Waiting for the process ~a quits...~%" pid)

	   (let loop ()
		 (condition-case (begin
						   (process-signal pid 0)
						   ;; まだ生きてゐるので 0.1 秒待ってから再試行
						   (thread-sleep! 0.1)
						   (loop))
	        [exn () #f])))))


(receive (options operands)
     (args:parse (command-line-arguments) opts)

   (with-output-to-port (current-error-port)
	 (lambda ()
	   (print "")
	   (print ".::. CleiteD - server-based RSS aggregator .::.")
	   (print "")

	   (define opt:uid (let [(info (user-information opt:user))]
						 (if info
							 (list-ref info 2)
							 (begin
							   (printf "The user ~a doesn't exist.~%" opt:user)
							   (exit 1)))))

	   (define opt:gid (let [(info (group-information opt:group))]
						 (if info
							 (list-ref info 2)
							 (begin
							   (printf "The group ~a doesn't exist.~%" opt:group)
							   (exit 1)))))
	   
	   ;; 先づ以下の全ての條件を滿さなければならない。
	   ;; [1] home directory が存在する事
	   ;;       ==> 無ければ作る
	   ;; [2] その所有者が指定された uid / gid に等しい事
	   ;;       ==> 違ふなら變へる
	   ;; [3] その permission が 700 である事
	   ;;       ==> 違ふなら變へる
	   (unless (directory? opt:home)
		 (print "The home directory of the CleiteD doesn't exist:")
		 (print "  ==> directory: " opt:home)
		 (create-directory opt:home)
		 (print "  ==> created"))

	   (let [(stat (file-stat opt:home))]
		 (unless (and (= opt:uid (vector-ref stat 3))
					  (= opt:gid (vector-ref stat 4)))
		   (printf "The owner of the home directory is not ~a:~a.~%"
				   opt:user opt:group)
		   (change-file-owner opt:home opt:uid opt:gid)
		   (print "  ==> changed")))

	   (let* [(mode (file-permissions opt:home))
			  (on?  (lambda (flag) (not (= 0 (bitwise-and mode flag)))))
			  (off? (lambda (flag)      (= 0 (bitwise-and mode flag))))]
		 
		 (unless (and (on?  perm/irusr) (on?  perm/iwusr) (on?  perm/ixusr)
					  (off? perm/irgrp) (off? perm/iwgrp) (off? perm/ixgrp)
					  (off? perm/iroth) (off? perm/iwoth) (off? perm/ixoth))
		   (printf "The permission of the home directory is not 700.~%" opt:home)
		   (change-file-mode opt:home (bitwise-ior perm/irusr perm/iwusr perm/ixusr))
		   (print "  ==> changed")))

	   ;; 現在の uid と gid を確認。指定されたものと異なってゐれば、そ
	   ;; れに變更する。變更するには root である必要がある。
	   (unless (and (= (current-effective-user-id)  opt:uid)
					(= (current-effective-group-id) opt:gid))
		 (printf "My effective UID and GID are not ~a:~a.~%" opt:user opt:group)
		 (set-group-id! opt:gid)
		 (set-user-id!  opt:uid)
		 (print "  ==> changed"))
		 
	   ;; opt:action を見て分岐
	   (cond ((eq? 'start opt:action)
			  (printf "CleiteD is going to daemonize...~%")
			  (daemonize-node))

			 ((eq? 'no-fork opt:action)
			  (start-node))

			 ((eq? 'stop opt:action)
			  (stop-node force: #f))

			 ((eq? 'kill opt:action)
			  (stop-node force: #t))

			 ((eq? 'restart opt:action)
			  (stop-node  force: #f)
			  (printf "CleiteD is going to daemonize...~%")
			  (daemonize-node))

			 (else
			  (error (sprintf "Unknown action: ~a" opt:action)))))))
