;; -*- hen -*-
(declare (unit opml)
		 (uses cleited-passive-node
			   cleited-utils
			   account
			   LWS-request
			   LWS-response
			   LWS-UA)
		 (export opml:install-methods!))

(use tinyclos
	 sqlite3
	 ssax
     sxml-tools
     srfi-13)


(define (opml:install-methods! node)
  (assert (instance-of? node <cleited-passive-node>))

  (wrap-method! node 'opml:parse  opml:parse)
  (wrap-method! node 'opml:export opml:export))



(define (opml:parse-from-port port)
  ;; SSAX は使ひ易いね。
  ((SSAX:make-parser NEW-LEVEL-SEED
					 (lambda (name attrs ns expect seed)
					   ;; この outline に xmlUrl 屬性が存在するならば、
					   ;; seed にレコードを付け加へる。
					   (if (eq? name 'outline)
						   (let [(xmlUrl-cell (assq 'xmlUrl attrs))
								 (title-cell  (assq 'title  attrs))]

							 (if xmlUrl-cell
								 `(,@seed
								   ((feed-uri . ,(cdr xmlUrl-cell))
									,@(if title-cell
										  `((title . ,(cdr title-cell)))
										  '())))
								 seed))
						   seed))

					 FINISH-ELEMENT
					 (lambda (name attrs ns parent-seed seed)
					   seed)

					 CHAR-DATA-HANDLER
					 (lambda (s1 s2 seed)
					   seed))
   port '()))


(define (opml:parse arg db-pod)
  (let [(opml (get-arg arg 'opml check: string? optional: #t))
		(uri  (get-arg arg 'uri  check: string? optional: #t))]

	(cond [(and opml (not uri))
		   (opml:parse-from-port
		       (open-input-string (util:xml->utf8-xml opml)))]

		  [(and uri (not opml))
		   (let* [(ua  (make <lws:UA>))
				  (req (make <lws:request> 'GET uri))
				  (res (request ua req))]
			 
			 ;; 200 の場合は OPML をパース。
			 (cond [(= 200 (lws:code-of res))
					(opml:parse-from-port
					    (open-input-string (util:xml->utf8-xml
											   (lws:content-of res))))]

				   [else
					(error `(transport-error . ,(lws:code-of res)))]))]

		  [else
		   (error "Parameter error")])))


(define (opml:export arg db-pod)
  (let [(auth (get-arg arg 'auth check: list?))]

    (account:check auth db-pod)

    (define user-name
      (db-pod:run db-pod
          (lambda (db)
            (sqlite3:first-result
             db (string-append "SELECT name       "
                               "  FROM accounts   "
                               " WHERE user_id = ?")
             (get-arg auth 'user-id)))))

    (define feed-list ;; (("Title" . "URL"), ...)
      (db-pod:run db-pod
          (lambda (db)
            (sqlite3:map-row
                (lambda (feed-uri title)
                  (cons feed-uri title))
                db
                (string-append "SELECT                        "
                               "    F.feed_uri,               "
                               "    F.title                   "
                               "FROM                          "
                               "    subscription as S         "
                               "LEFT JOIN                     "
                               "    feeds AS F USING (feed_id)"
                               "WHERE                         "
                               "    S.user_id = ?             ")
                (get-arg auth 'user-id)))))

    (util:cat-normal
     (util:sxml->xml `(opml
                       (@ (version "1.0"))

                       (head
                        (title "Cleite - Feed list for " ,user-name))

                       (body
                        ,(map (lambda (feed)
                                `(outline
                                  (@ (title  ,(cdr feed))
                                     (xmlUrl ,(car feed)))))
                              feed-list))))
     "<?xml version=\"1.0\" encoding=\"UTF-8\"?>")))
