;; -*- hen -*-
(declare (unit throttle)
		 (export <throttle>
				 initialize
				 throttle:enqueue))

(use tinyclos
	 srfi-18)

(define-class <throttle> ()
  (max-concurrency     ;; 最大並列動作スレッド數 :: <integer>
   current-concurrency ;; 現在の動作中スレッド數 :: <integer>
   queue               ;; 動作待ちスレッド        :: <list>
   mutex               ;; キューの爲の mutex
   dequeued            ;; スレッド終了時に signal! される條件變數
   ))

(define-method (initialize (this <throttle>) args)
  (call-next-method)

  (define (init #!key concurrency)
	(assert (positive? concurrency))

	(initialize-slots this
					  (list 'max-concurrency     concurrency
							'current-concurrency 0
							'queue               (list)
							'mutex               (make-mutex)
							'dequeued            (make-condition-variable))))
  (apply init args))

(define-method (throttle:enqueue (this  <throttle>)
								 (thunk <procedure>))

  (define mutex    (slot-ref this 'mutex))
  (define dequeued (slot-ref this 'dequeued))
  
  (dynamic-wind
	  (lambda ()
		(mutex-lock! mutex))

	  (lambda ()
		;; スレッド動作數に余裕が無ければ、余裕が出來るまで待つ。
		(let loop ()
		  (when (>= (slot-ref this 'current-concurrency)
					(slot-ref this 'max-concurrency))
			(mutex-unlock! mutex dequeued)
			(mutex-lock!   mutex)
			(loop)))

		;; 動作開始。動作中はロックを外して置く。
		(dynamic-wind
			(lambda ()
			  (slot-set! this 'current-concurrency
						 (+ 1 (slot-ref this 'current-concurrency)))
			  (mutex-unlock! mutex))

			(lambda ()
			  (thunk))

			(lambda ()
			  (mutex-lock! mutex)
			  (slot-set! this 'current-concurrency
						 (- (slot-ref this 'current-concurrency) 1)))))

	  (lambda ()
		(mutex-unlock! mutex)
		(condition-variable-signal! dequeued))))
