# -*- sh -*-
dnl
dnl AM_CHECK_FOR_EGG(
dnl     EGG-NAME [, MINIMUM-VER
dnl              [, ACTION-IF-FOUND
dnl              [, ACTION-IF-NOT-FOUND]]])
dnl
AC_DEFUN([AM_CHECK_FOR_EGG],
	[
		if test "$CSI" = ""; then
			CSI=csi
		fi

		egg_checker="
			(match-define (\"-e\" _ \"--\" egg min-ver) (command-line-arguments))
			(define info (extension-information (string->symbol egg)))
            
            (unless info
                (exit 1))

            (define ver
                (if (assq 'version info)
                    (cadr (assq 'version info))
                    #f))

            (if ver
                (printf \"(~s) \" ver))
                       

            (if (equal? min-ver \"\")
                ;; min-ver is not specified
                (exit 0)

                (if (>= ver (string->number min-ver))
                    (exit 0)
                    (exit 1)))
"

		egg_loader="
            (match-define (\"-e\" _ \"--\" egg) (command-line-arguments))
            (eval \`(use ,(string->symbol egg)))
"

		# is the egg installed?

		if test "$2" = ""; then
			AC_MSG_CHECKING([if $1's egg is installed])
		else
			AC_MSG_CHECKING([if $1's egg >= $2 is installed])
		fi

		egg_is_installed=

		if $CSI -e "$egg_checker" -- "$1" "$2"; then
			AC_MSG_RESULT([yes])
			egg_is_installed=yes
		else
			AC_MSG_RESULT([no])
			egg_is_installed=no
		fi

		# is the egg loadable?

		egg_is_loadable=

		if test "$egg_is_installed" = "yes"; then
			AC_MSG_CHECKING([if $1's egg is actually loadable])

            egg_module=$1

			if $CSI -e "$egg_loader" -- "$egg_module"; then
				AC_MSG_RESULT([yes])
				egg_is_loadable=yes
			else
				AC_MSG_RESULT([no])
				egg_is_loadable=no
			fi
		fi

		# do the action
		if test "$egg_is_installed" = "yes" \
			-a "$egg_is_loadable" = "yes"; then

			ifelse([$3],, :, [$3])
		else
			ifelse([$4],, :, [$4])
		fi
		])

dnl
dnl AM_ASSERT_EGG_AVAILABLE(
dnl     EGG-NAME [, MINIMUM-VER])
dnl
AC_DEFUN([AM_ASSERT_EGG_AVAILABLE],
	[
		AM_CHECK_FOR_EGG(
			[$1], [$2],
			[],
			[echo "***"
             echo "*** $1's egg is not available."
             echo "*** It may be possible to be installed with the following command:"
             echo "***"
             echo "*** % chicken-setup $1"
             echo "***"
             exit 1])
		])